#include "buffer.h"
#include "util.h"
#include <stdlib.h>
#include <string.h>
#include "compat.h"


/** vraci true, je li znak v bufferu */
int have_char(char c,char *buf,int len) {
 int i;
 for (i=0;i<len;i++) {
  if (buf[i]==c) return 1;
 }
 return 0;
}

/** inicializuje promennou na defaultni hodnoty + naalokuje pocatecni buffer */
void buf_init(charbuffer *b,int size) {
 b->buf=malloc1(size);
 b->size=0;
 b->a_size=size;
}

#define BUFFER_GRANULARITY 4096//granularita bufferu, musi byt mocnina dvou

/** realokuje buffer na velikost alespon minsize bajtu. Pokud je buffer uz dostatecne velky, nedela nic */
void buf_realloc(charbuffer *b,int minsize) {
 char *nb;
 if (minsize>b->a_size) {
  b->a_size=(minsize+(BUFFER_GRANULARITY-1)) & (0-BUFFER_GRANULARITY);//realokuje na nejblizsi vyssi nasobek 512 bajtu
  #ifdef DEBUG
  if (minsize>b->a_size) fatalf("buf_add: sem se splet .... %d,%d",minsize,b->a_size);
  #endif
  nb=realloc(b->buf,b->a_size);
  if (!nb) fatalf("Failed to reallocate buffer to %d bytes",b->a_size);
  b->buf=nb;
 }
}

/** prida do bufferu data */
void buf_add(charbuffer *b,const char *data,int datalen) {
 if (!datalen) return;
 buf_realloc(b,b->size+datalen);
 memcpy(b->buf+b->size,data,datalen);
 b->size+=datalen;
}

/** uvolni pamet zabranou bufferem */
void buf_free(charbuffer *b) {
#ifdef DEBUG
 if (!(b->buf)) {
  b=0;
  b->buf=0;//aby to vyvolalo segfault a ja pk moh pitvat CORE
  fatalf("freeing buffer twice!");
 }
#endif
 free(b->buf);
#ifdef DEBUG
 b->buf=0;
#endif
}

/** vrati a odstrani radky ukoncenou CRLF z bufferu, nebo NULL pokud tam takova radka neni */
char *buf_getline(charbuffer *b) {
 int i;
 char *u=b->buf;
 char *res;
 if (b->size<2) return NULL;
 for (i=1;i<b->size;i++) {//od 2.bajtu
  if (u[i]=='\n' && u[i-1]=='\r') { //CRLF
   u[i-1]=0;
   res=strdup(u);
   b->size-=(i+1);
   memmove(u,u+i+1,b->size);
   return res;
  }
 }
 return NULL;
}

/** vrati a radku ukoncenou CRLF, LF, CR nebo koncem bufferu z bufferu, nebo NULL pokud je buffer prazdny.
 Radku z bufferu neostrani, buffer zustane nezmenen */
char *buf_firstline(charbuffer *b) {
 int i;
 char *u=b->buf;
 char *res;
 if (b->size==0) return NULL;
 for (i=0;i<b->size;i++) {//od 1.bajtu
  if (u[i]=='\n' || u[i]=='\r') { //CR nebo LF
   res=malloc1(i+1);
   memmove(res,b->buf,i);
   res[i]=0;
   return res;
  }
 }
 //neni konec radky, vratim cely buffer
 res=malloc1(b->size+1);
 memmove(res,b->buf,b->size);
 res[b->size]=0;
 return res;
}

static const char *crlf="\r\n";//retezec CR+LF

/** prida do buffer radku line + znaky CR a LF */
void buf_addline(charbuffer *b,const char *line) {
 buf_add(b,line,strlen(line));
 buf_add(b,crlf,2);
}
