#include "config.h"
#include "cfg_file.h"
#include <pthread.h>
#include <stdio.h>
#include <string.h>
#include "compat.h"

config *active; //aktivni konfigurace [EXT]
char *configfile;//jmeno aktualniho konfiguracniho souboru [EXT]

/** nahraje do stromu vsechny radky ze souboru krome komentaru.
 Pripousti se, ze soubor neexistuje */
tree *loadlist(char *filename) {
 int list_size;//velikost
 char *list_data;//nacteno v pameti
 char *list_line;//Vyparsovana radka
 text_token list;
 tree *t=NULL;
 debug_printf("%s try loading\n",filename);
 list_size=loadfile((void **)&list_data,filename);
 if (!list_data) return NULL;
 debug_printf("%s loading\n",filename);

 list=token_create(list_data,list_size);
 while((list_line=token_readline(&list))) {
  debug_printf("[%s] %s\n",filename,list_line);
  tree_add(&t,list_line,(void *)1);//je jedno co se tam da, neni dulezita hodnota, ale to ze tam prvek je
  free(list_line);
 }
 debug_printf("%s loaded\n",filename);
 free(list_data);
 return t;
}

/** nahraje konfiguraci ze souboru file */
config *load_config(char *filename) {
 config *cfg;//konfiguracni data
 int cfg_size;//velikost konfiguraku
 char *cfg_data;//konfigurak nacteny v pameti
 char *cfg_key;//Vyparsovany klic
 char *cfg_value;//Vyparsovany klic
 text_token conf;//konfigurak v pameti v parsovatelnem tokenu

 //alokuj konfiguraci, nastav default data
 cfg=malloc0(sizeof(config));
 cfg->pop3_port=110;
 cfg->smtp_port=25;
 cfg->refcount=0;
 cfg->blacklist_level=100;
 cfg->whitelist_level=0;

 //nahraj soubor, zacni parsovat
 cfg_size=loadfile((void **)(&cfg_data),filename);
 if (!cfg_data) fatalf("Can't load config file %s",filename);
 if (!cfg_size) fatalf("Empty config file %s",filename);
 conf=token_create(cfg_data,cfg_size);
 while(token_readkv(&conf,&cfg_key,&cfg_value)) {
  debug_printf("KEY: %s =  VALUE: %s\n",cfg_key,cfg_value);//debug
       if (eqi(cfg_key,"pop3_port"))		cfg->pop3_port=atoi(cfg_value);
  else if (eqi(cfg_key,"smtp_port"))		cfg->smtp_port=atoi(cfg_value);
  else if (eqi(cfg_key,"blacklist-level"))	cfg->blacklist_level=atoi(cfg_value);
  else if (eqi(cfg_key,"whitelist-level"))	cfg->whitelist_level=atoi(cfg_value);
  else if (eqi(cfg_key,"classification"))	cfg->p_class=strdup(cfg_value);
  else if (eqi(cfg_key,"smtp_server"))		cfg->remotesmtp=strdup(cfg_value);
  else if (eqi(cfg_key,"blacklist"))		cfg->blacklistfile=strdup(cfg_value);
  else if (eqi(cfg_key,"whitelist"))		cfg->whitelistfile=strdup(cfg_value);
  else if (eqi(cfg_key,"post-action"))		plist_add(&cfg->p_postact,cfg_value);
  else if (eqi(cfg_key,"post-filter"))		plist_add(&cfg->p_postfilt,cfg_value);
  else if (eqi(cfg_key,"classify-filter"))	plist_add_num(&cfg->p_prefilt,0,0,cfg_value);
  else if (eqi(cfg_key,"mark-subject"))		plist_add(&cfg->subject_mark,cfg_value);
  else fatalf("Unknown key in config file: %s",cfg_key);
  free(cfg_key);//uvolni i cfg_value
 }
 free(cfg_data);

 debug_printf ("loading lists\n");
 //nahrani blacklistu/whitelistu
 if (!empty(cfg->blacklistfile)) cfg->blacklist=loadlist(cfg->blacklistfile);
 if (!empty(cfg->whitelistfile)) cfg->whitelist=loadlist(cfg->whitelistfile);
 debug_printf ("loaded lists\n");

 lock_config_int(cfg);//bez zamykani (tady ani nema smysl, pac ta struktura jeste nevysla ven)
 return cfg;
}

pthread_mutex_t cfg_lock;//lock na manipulace s configem

/** inicializace zamku */
void config_mutex_lock_init() {
 pthread_mutex_init(&cfg_lock, NULL);
}

/** zamkne konfiguraci pro modifikace */
void config_mutex_lock() {
 //printf("c_lock\n");
 pthread_mutex_lock(&cfg_lock);
}

/** odemkne konfiguraci pro modifikace */
void config_mutex_unlock() {
 //printf("c_unlock\n");
 pthread_mutex_unlock(&cfg_lock);
}


/** zvysi refcount */
config *lock_config(config *c) {
 config_mutex_lock();
 if (!c) c=active;
 c->refcount++;
 config_mutex_unlock();
 return c;
}

/** zvysi refcount (bez zamykani) */
config *lock_config_int(config *c) {
 if (!c) c=active;
 c->refcount++;
 return c;
}

/** snizi refcount, eventuelne zlikviduje strukturu */
void unlock_config(config *c) {
 config_mutex_lock();
 if (!c) c=active;
 unlock_config_int(c);
 config_mutex_unlock();
}

/** snizi refcount, eventuelne zlikviduje strukturu (bez zamykani) */
void unlock_config_int(config *c) {
 c->refcount--;
 if (!c->refcount) {
  free_config(c);
 } 
}

void main_rebind(int change_smtp,int change_pop3);//funkce v hlavnim programu, ktera zmeni porty na kterych se nasloucha

/** sighup handler: reload config */
void sighup_handler(int sig) {
 int smtpp,pop3p;//stare porty
 config *cfg;
 //printf("sigHUP received\n");
 config_mutex_lock();
 cfg=load_config(configfile);
 smtpp=active->smtp_port;
 pop3p=active->pop3_port;
 unlock_config_int(active);
 //printf("cfg: %x->%x\n",(int)active,(int)cfg);
 active=cfg;
 if (smtpp!=active->smtp_port || pop3p!=active->pop3_port) {
  //prehazi sockety v hlavnim programu, pokud se listenuje na novych portech ....
  main_rebind(smtpp!=active->smtp_port,pop3p!=active->pop3_port);
 }
 config_mutex_unlock();
 //printf("sigHUP finished\n");
}

/** uvolni konfiguracni strukturu z pameti*/
void free_config(config *c) {
 str_free(c->blacklistfile);
 str_free(c->whitelistfile);
 tree_free(c->blacklist);
 tree_free(c->whitelist);
 str_free(c->p_class);
 plist_free(c->p_prefilt);
 plist_free(c->p_postfilt);
 plist_free(c->p_postact);
 plist_free(c->subject_mark);
 free(c);
}

/** odalokuje spojak s programy spoustenymi dle priorit */
void plist_free(priority_list *p) {
 priority_list *px;
 while (p) {
  px=p->next;
  free(p->prog);
  free(p);
  p=px;
 }
}

/** do prioritniho seznamu prida zaznam z konfiguraku (format "LOW-HIGH PROGRAM"/"PRIO PROGRAM") */
void plist_add(priority_list **p,char *cfg_line) {
 int lo,hi,res;
 char *prog;
 res=sscanf(cfg_line,"%d-%d ",&lo,&hi);
 if (res!=2) {
  res=sscanf(cfg_line,"%d ",&lo);
  if (res==1) hi=lo;
  else fatalf("Invalid/missing priority in priority list (%s)",cfg_line);
 }
 prog=cfg_line;
 while (prog[0] && (!is_space(prog[0]))) prog++;
 while (prog[0] && (is_space(prog[0]))) prog++;//preskocit taky mezery
 if (empty(prog)) fatalf("Empty program in priority list (%s)",cfg_line);
 plist_add_num(p,lo,hi,prog);
}

/** do prioritniho seznamu prida zaznam z konfiguraku (LOW,HIGH,PROGRAM) */
void plist_add_num(priority_list **p,int lo, int hi, char *prog) {
 priority_list *px;
 if (empty(prog)) fatalf("Empty program in program list");
 while(*p) p=&((*p)->next);//postoupime na posledni zaznam
 px=malloc1(sizeof(priority_list));
 px->next=NULL;
 px->min=lo;
 px->max=hi;
 px->prog=strdup(prog);
 *p=px;//pridame na konec seznamu
}
