/***
 ruzne funkce:

 prace s binarnimi vyhledavacimi stromy
 prace s tokeny
 + dalsi
*/
#include "util.h"
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <math.h>
#include "compat.h"


/** vrati velikost souboru */
int fsize(const char *f) {
 struct stat b;
 if (stat(f,&b)) return 0;//return 0 if error
 return b.st_size;
}

/** nahraje soubor do pameti, vrati velikost */
int loadfile(void **out,const char *name) {
 int f;
 int s=0,readb,totalb=0;
 f=open(name,0);
 if (f==-1) {
  *out=0;
  return 0;
 }
 s=fsize(name);
 if (s) {
  *out=malloc(s);
  if (*out) {
   while (totalb<s) {
    readb=read(f,(*out)+totalb,s);
    totalb+=readb;
    if (readb==0) { //soubor se nepodarilo z jakehosi zahadneho duvodu (cygwin?) nahrat cely -> usekneme
     if (totalb==0) { 
      free(*out);
      (*out)=(void *)0xdeadbeef;
     } else { 
      *out=realloc(*out,totalb);
     }
     break;
    }
   }
  }
 } else (*out)=(void *)0xdeadbeef;//"naalokovany pointer o velikosti 0" -> stejne na nej nejde hrabnout, staci kdyz neni null
 (void)close(f);
 return totalb;
}

/** vyskoci s chybovou hlaskou. Dobre k debugovani */
void fatal(char *s){
 printf("%s\n",s);
 exit(EXIT_FAILURE);
}

/** jako fatal(), ale s formatovanym retezcem jako u printf */
void fatalf(char *fmt,...){
 va_list argptr;
 char s[256];
 va_start(argptr,fmt);	
 _vsnprintf(s,sizeof(s),fmt,argptr);
 va_end(argptr);
 fatal(s);
}

/** odalokuje strom z pameti */
void tree_free(tree *t) {
 if (!t) return;
 tree_free(t->r);
 tree_free(t->l);
 free(t->name);
 free(t);
}

/** zduplikuje dany strom - vytvori novou kopii */
tree *tree_dup(const tree *t) {
 tree *nt=0;
 if (!t) return 0;
 tree_add(&nt,t->name,t->val);
 nt->num=t->num;
 nt->l=tree_dup(t->l);
 nt->r=tree_dup(t->r);
 return nt;
}

/** vytahne ze stromu ukazatel na list s prvek s cislem i (i je mezi 0 a (poctem prvku ve stromu -1)
Da se pouzit napr. pro prochazeni cleeho stropu, ackoliv tree_foreach / tree_foreach_p muze byt vhodnejsi
Pokud nastane chyba nebo je prvek mimo rozsah, vrati NULL
 */
const tree *tree_element(const tree *t,int i) {
 if (!t) return 0;
 if (i==0) return t;//nasli jsme!
 if (i>=t->num) return 0;//tolik prvku ve strome neni
 if (!t->l) return tree_element(t->r,i-1);
 if (i<=t->l->num) {//je to v leve casti
  return tree_element(t->l,i-1);
 } else { //je to v prave casti (nebo mimo)
  return tree_element(t->r,i-1-t->l->num);
 }
}

/** prohleda strom jestli obsahuje polozku name a vrati ji, pokud neobsahuje, vraci 0 */
void *tree_search(const tree *t,const char *name) {
 int r;
 if (!t) return 0;
 if (!name) return 0;//NULL pointer tam fakt nebude
 r=stricmp(name,t->name);
 if (r>0) return tree_search(t->l,name);
 if (r<0) return tree_search(t->r,name);
 return t->val;
}

/** prida do stromu polozku se jmenem name a hodnovou val1, pokud je jiz obsazena, neprida nic (a pokud je zapnuty debug tak rovnou fatalne spadne) */
void tree_add(tree **t,const char *name,void *val1) {
 int r;
 if (!name) return;//NULL pointer tam pridavat odmitam
 if (!*t) {//root==null
  *t=(tree *)malloc0(sizeof(tree));
 } else { //left or right
  r=stricmp((*t)->name,name);
  (*t)->num++;
  if (r<0) {
   tree_add(&((*t)->l),name,val1);
   return;
  }
  if (r>0){
   tree_add(&((*t)->r),name,val1);
   return;
  }
// #ifdef DEBUG /// mno tohle se tu nehodi ... pokud o vlozeni duplicit je bezny
//  fatalf("assert: tree_add (%s): already present",name);
// #endif
  return;
 }
 (*t)->num++;
 (*t)->name=strdup(name);
 (*t)->val=val1;
}

/** zavola funkci na kazdy prvek stromu */
void tree_foreach(tree *t,tree_func *f) {
 if (!t) return;
 tree_foreach(t->r,f);
 f(t->name,t->val);
 tree_foreach(t->l,f);
}

/** zavola funkci na kazdy prvek stromu (s extra parametrem)*/
void tree_foreach_p(tree *t,tree_func_p *f,void *param) {
 if (!t) return;
 tree_foreach_p(t->r,f,param);
 f(t->name,t->val,param);
 tree_foreach_p(t->l,f,param);
}

/** odstrani ze stormu polozku name
 1 - polozka byla odstranena
 0 - polozka tam nebyla
 */
int tree_del(tree **t,const char *name) {
 int r;
 tree *op,*op2,*op3;
 if (!*t) return 0;//neni co mazat
 r=stricmp((*t)->name,name);
 if (r<0) {
  return tree_del(&((*t)->l),name);
 }
 if (r>0) {
  return tree_del(&((*t)->r),name);;
 }
 //trefa!
 op=*t;
 if (!op->l) {
  *t=op->r;
 } else if (!op->r) {
  *t=op->l;
 } else {//oba podstromy.
  *t=op->r;
  op3=op->l;//druha cast stromu
  op2=op->r;//sem ji chci zavesi
  while (op2->l) op2=op2->l;
  op2->l=op3;
 }
 free(op->name);
 free(op);
 return 1;
}

/** odstrani ze stormu polozku name a vrati ji (nebo NULL, pokud tam neni)
 1 - polozka byla odstranena
 0 - polozka tam nebyla
 */
void *tree_delv(tree **t,const char *name) {
 int r;
 tree *op,*op2,*op3;
 void *o;
 if (!*t) return 0;//neni co mazat
 r=stricmp((*t)->name,name);
 if (r<0) {
  return tree_delv(&((*t)->l),name);
 }
 if (r>0) {
  return tree_delv(&((*t)->r),name);;
 }
 //trefa!
 op=*t;
 if (!op->l) {
  *t=op->r;
 } else if (!op->r) {
  *t=op->l;
 } else {//oba podstromy.
  *t=op->r;
  op2=op->r->l;
  op3=op->l;
  op->r->l=op3;
  while (op3->r) op3=op3->r;
  op2=op3->r;
 }
 free(op->name);
 o=op->val;
 free(op);
 return o;
}

/** vrati pocet prvku stromu */
int tree_count(tree *t) {
 if (t) return t->num;
 return 0;
}

// funkce na praci s textovymi tokeny

/** preskoci vse v tokenu az do konce radky */
void token_remove_rest_of_line(text_token *t) {
 while (t->left && !newline(t->curr[0])) {
  token_move(t,1);
 }
}

/** preskoci vsechny whitespacy a pokud narazi na komentar, tak i ten */
void token_skip_trash(text_token *t) {
 while (t->left) {// je li neco v bufferu
  while (token_whitespace(t)) {//skip whitespace
   token_move(t,1);
  }
  if (t->left>=2) {
   if (t->curr[0]==t->curr[1] && t->curr[1]=='/') {
    token_remove_rest_of_line(t);
    continue;// za komentarem muze byt nekolik newlinu
   }
  }
  if (t->left>=1) {
   if (t->curr[0]=='#') {
    token_remove_rest_of_line(t);
    continue;// za komentarem muze byt nekolik newlinu
   }
  }
  return;
 }
}

/** nacte znak z tokenu a odstrani ho. Vraci 1 pri uspechu, 0 pri chybe (je tam jiny znak, nebo tam neni zadny) */
int token_readchar(text_token *t,char c) {
 token_skip_trash(t);
 if (token_beginwith(t,c)) {
  token_move(t,1);
  return 1;
 }
 return 0;
}

/** nacte retezec v uvozovkach z tokenu */
char *token_readqstring(text_token *t) {
 char *begin,*ret;
 int len;
 token_skip_trash(t);
 token_readchar(t,'"');
 begin=t->curr;
 while (!token_beginwith(t,'"')) {
  if (t->left==0) return 0;// chyba, neukonceny retezec
  token_move(t,1);
 }
 len=t->curr-begin;
 ret=malloc1(len+1);
 b_strncpy(ret,begin,len+1);
 token_readchar(t,'"');
 return ret;
}

/** nacte z tokenu kus binarnich dat */
int token_loadbinary(text_token *t,void *i,int s) {
 if (t->left<s) return 0;// chyba, malo znaku
 memmove(i,t->curr,s);
 token_move(t,s);
 return 1;
}

/** nacte retezec z tokenu (znaky az do prvni whitespace nebo konce radky) */
char *token_readstring(text_token *t) {
 char *begin,*ret;
 int len;
 token_skip_trash(t);
 begin=t->curr;
 while (!token_whitespace(t)) {
  if (t->left==0) return 0;// chyba, neukonceny retezec
  token_move(t,1);
 }
 len=t->curr-begin;
 ret=malloc1(len+1);
 b_strncpy(ret,begin,len+1);
 return ret;
}

/** ze zacatku tokenu nacte a odstrani zaznam ve formatu key=value.
 Misto = lze pouzit i jiny separator
 POZOR! dealokovat se musi jenom key, nikdy ne value, nebot to je jeden buffer */ 
int token_readkv_sep(text_token *t,char **key,char **value,char separator) {
 int end;
 char *line,*k,*l;
 for (;;) { //cte tak dlouho dokud nenacte platnou radku, nebo nedojde na konec tokenu
  line=token_readline(t);
  if (!line) return 0;
  *key=line;
  k=strchr(line,separator);
  if (k) {
   *k=0;
   l=k+1;
   while (k>*key && isseparator(k[-1])) {//odstran whitespace pred =
    k--;
    *k=0;
   }
   while (isseparator(*l)) l++;//whitespace za =
   end=strlen(l);//odstranit whitespace na konci radky
   while (end>0 && is_space(l[end-1])) {
    l[end-1]=0;
    end--;
   }
   *value=l;
   return 1;
  }
  free(line);
 }
}

/** nacte z tokenu jeden float */
int token_readfloat(text_token *t,double *f) {
 char *s;
 token_skip_trash(t);
 s=token_readstring(t);
 if (!s) return 0;
 *f=atof(s);
 free(s);
 return 1;
}

/** nacte z tokenu jeden integer */
int token_readint(text_token *t,int *f) {
 char *s;
 token_skip_trash(t);
 s=token_readstring(t);
 if (!s) return 0;
 *f=atoi(s);
 free(s);
 return 1;
}


/** uvolni token z pameti, vcetne jeho vnitrku, tedy dat, ze kterych byl token vytvoren */
void freetoken(text_token *t) {
 free(t->begin);
 free(t);
}

/** nacte z tokenu jednu neprazdnou radku */
char *token_readline(text_token *t) {
 int i=0;
 char *ret;
 token_skip_trash(t);
 if (!t->left) return 0;
 while (t->left && !eol(t->curr[i])) {
  if (t->curr[i]=='/' && t->curr[i+1]=='/') break;
  if (t->curr[i]=='#') break;
  t->left--;
  i++;
 }
 ret=malloc1(i+1);
 memmove(ret,t->curr,i);
 ret[i]=0;
 t->curr+=i;
 if (t->curr[0]=='/' && t->curr[1]=='/') token_remove_rest_of_line(t);//komentar
 if (t->curr[0]=='#') token_remove_rest_of_line(t);//komentar
 while (t->left && eol(t->curr[0])) {//odstranit konec radku, preskocit pripadne prazdne radky
  t->left--;
  t->curr++;
 }
 return ret;
}

/** vytvori z danych dat parsovatelny token */
text_token token_create(void *source,int size) {
 text_token x;
 x.curr=x.begin=source;
 x.left=size;
 return x;
}

/** naalokuje pamet. Selhani neni povoleno (pri nedostatku pameti program konci s chybovou hlaskou) */
void *malloc1(size_t t){
 void *c;
 c=malloc(t);
 if (!c) fatalf("Out of memory (%d bytes)",t);
 return c;
}

/** naalokuje pamet. Pamet je vyplnena nulama */
void *malloc0(size_t t){
 void *c;
 c=malloc1(t);
 memset(c,0,t);
 return c;
}

/** jako strncpy, ale posledni znak vzdy nastavi na NULL */
char *b_strncpy(char *dest,const char *source, size_t count) {
 char *q=strncpy(dest,source,count);
 q[count-1]=0;
 return q;
}

/** vrati 1 pokud se retezce rovnaji */
int eq(const char *a,const char *b) {
 if (empty(a)) return (empty(b));
 if (empty(b)) return 0;
 return (strcmp(a,b)==0);
}

/** vrati 1 pokud se retezce rovnaji (case insensitive) */
int eqi(const char *a,const char *b) {
 if (empty(a)) return (empty(b));
 if (empty(b)) return 0;
 return (stricmp(a,b)==0);
}

/** odalokace retezce s kontrolou na NULL pointery */
void str_free(char *s) {
 if (s==NULL) return;
 free(s);
}

/** vrati 1 pokud je retezec NULL nebo prazdny */
int empty(const char *a) {
 if (!a) return 1;
 return *a==0;
}

/* * duplikace retezce */
/*char *strdup(const char *s){
 size_t i=strlen(s);
 char *t=malloc1(i+1);
 strcpy(t,s+1);
// t[i]=0;
 return t;
}*/

/** vrati 1 pokud retezec s2 zacina retezcem s1 (case insensitive), jinak 0 */
int str_beginwithi(const char *s1,const char *s2) {
 int s1l=strlen(s1);
 int s2l=strlen(s2);
 if (s1l>s2l) return 0;
 return strnicmp(s1,s2,s1l)==0;
}

/** pokud retcode je -1, vypise perror s parametrem action a ukonci program
 typicke pouziti: chkerror("popis chyby",syscall_co_vraci_-1_pri_chybe(...));
*/
int chkerror(char *action,int retcode) {
 if (retcode!=-1) return retcode;
 perror(action);
 exit(EXIT_FAILURE);
 return -1;
}
